/**
  *  \file planetscentral/format/specpacker.cc
  *  \brief Format - Specification Files
  */

#include "planetscentral/format/specpacker.h"
#include "planetscentral/dbclient/connection.h"
#include "game/struct.h"
#include "io/io.h"
#include "int/array.h"
#include "int/hash.h"

/** Unpack a TCost.
    \param cost [in] Data
    \return Hash containing the cost */
static IntValue*
unpackCost(const TCost& cost)
{
    Ptr<IntHashData> hd = new IntHashData();
    hd->setNew("MC", makeIntValue(cost.mc));
    hd->setNew("T",  makeIntValue(cost.t));
    hd->setNew("D",  makeIntValue(cost.d));
    hd->setNew("M",  makeIntValue(cost.m));
    return new IntHash(hd);
}

/** Pack a TCost.
    \param cost [out] Data
    \param pc [in] Hash provided by user */
static void
packCost(TCost& cost, IntValue* pc)
{
    cost.mc = toIntResult(getHashElem(pc, "MC"));
    cost.t  = toIntResult(getHashElem(pc, "T"));
    cost.d  = toIntResult(getHashElem(pc, "D"));
    cost.m  = toIntResult(getHashElem(pc, "M"));
}

/****************************** EnginePacker *****************************/

/*
  Engine().COST.D                         engine[].COST->D
  Engine().COST.M                         engine[].COST->M
  Engine().COST.MC                        engine[].COST->MC
  Engine().COST.STR                       -
  Engine().COST.T                         engine[].COST->T
  Engine().FUELFACTOR                     engine[].FUELFACTOR[]
  Engine().ID                             -
  Engine().NAME                           engine[].NAME
  Engine().NAME.SHORT                     -
  Engine().SPEED$                         engine[].SPEED -- not in packer
  Engine().TECH                           engine[].TECH
  Engine().TECH.ENGINE                    -
*/

string_t
EnginePacker::pack(IntValue* data)
{
    string_t result;
    for (uint32_t i = 0, n = getArraySize(data); i < n; ++i) {
        IntValue* p = getArrayElem(data, i);
        IntValue* pf = getHashElem(p, "FUELFACTOR");

        TEngine e;
        packString(e.name, sizeof(e.name), toStringResult(getHashElem(p, "NAME")));
        packCost(e.cost, getHashElem(p, "COST"));
        e.tech    = toIntResult(getHashElem(p, "TECH"));
        for (uint32_t sp = 0; sp < 9; ++sp) {
            e.fuel[sp] = toIntResult(getArrayElem(pf, sp+1));
        }

        storeStructure(result, e);
    }
    return result;
}

IntValue*
EnginePacker::unpack(const string_t& data)
{
    Ptr<IntArrayData> ad = new IntArrayData();
    TEngine e;
    string_t::size_type pos = 0;
    while (getStructure(data, pos, e)) {
        Ptr<IntHashData> hd = new IntHashData();
        hd->setNew("NAME", makeStringValue(unpackString(e.name, sizeof(e.name))));
        hd->setNew("COST", unpackCost(e.cost));
        hd->setNew("TECH", makeIntValue(e.tech));

        Ptr<IntArrayData> adf = new IntArrayData();
        adf->content.pushNew(makeIntValue(0));
        for (uint32_t sp = 0; sp < 9; ++sp) {
            adf->content.pushNew(makeIntValue(e.fuel[sp]));
        }
        adf->addDimension(adf->content.size());
        hd->setNew("FUELFACTOR", new IntArray(adf));

        ad->content.pushNew(new IntHash(hd));
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}

/***************************** TorpedoPacker *****************************/

/*
  Torpedo().COST.D                        torp[].TORPCOST->D
  Torpedo().COST.M                        torp[].TORPCOST->M
  Torpedo().COST.MC                       torp[].TORPCOST->MC
  Torpedo().COST.STR                      -
  Torpedo().COST.T                        torp[].TORPCOST->T
  Launcher().COST.D                       torp[].TUBECOST->D
  Launcher().COST.M                       torp[].TUBECOST->M
  Launcher().COST.MC                      torp[].TUBECOST->MC
  Launcher().COST.STR                     -
  Launcher().COST.T                       torp[].TUBECOST->T
  Torpedo().DAMAGE                        torp[].DAMAGE1            -- not doubled like in c2server, hence renamed!
  Torpedo().ID                            -
  Torpedo().KILL                          torp[].KILL1              -- not doubled like in c2server, hence renamed!
  Launcher().MASS                         torp[].MASS     of launcher
  Torpedo().NAME                          torp[].NAME
  Torpedo().NAME.SHORT                    -
  Torpedo().TECH                          torp[].TECH
  Torpedo().TECH.TORPEDO                  -
*/

string_t
TorpedoPacker::pack(IntValue* data)
{
    string_t result;
    for (uint32_t i = 0, n = getArraySize(data); i < n; ++i) {
        IntValue* p = getArrayElem(data, i);

        TTorpedo t;
        packString(t.name, sizeof(t.name), toStringResult(getHashElem(p, "NAME")));
        packCost(t.laun_cost, getHashElem(p, "TUBECOST"));
        t.torp_cost = toIntResult(getHashElem(getHashElem(p, "TORPCOST"), "MC"));
        t.laun_mass = toIntResult(getHashElem(p, "MASS"));
        t.tech      = toIntResult(getHashElem(p, "TECH"));
        t.kill      = toIntResult(getHashElem(p, "KILL1"));
        t.damage    = toIntResult(getHashElem(p, "DAMAGE1"));

        storeStructure(result, t);
    }
    return result;
}

IntValue*
TorpedoPacker::unpack(const string_t& data)
{
    Ptr<IntArrayData> ad = new IntArrayData();
    TTorpedo t;
    string_t::size_type pos = 0;

    /* A typical torpspec file has a few bytes at its end that do not correspond
       to a real torpedo. Thus, limit unpacking to ten elements, i.e. position
       less than end-of-tenth torp. */
    while (pos < NUM_TORPS*t.size && getStructure(data, pos, t)) {
        Ptr<IntHashData> hd = new IntHashData();
        hd->setNew("NAME", makeStringValue(unpackString(t.name, sizeof(t.name))));

        TCost torp_cost;
        torp_cost.mc = t.torp_cost;
        torp_cost.t = torp_cost.d = torp_cost.m = 1;
        hd->setNew("TORPCOST", unpackCost(torp_cost));
        hd->setNew("TUBECOST", unpackCost(t.laun_cost));
        hd->setNew("MASS",     makeIntValue(t.laun_mass));
        hd->setNew("TECH",     makeIntValue(t.tech));
        hd->setNew("KILL1",    makeIntValue(t.kill));
        hd->setNew("DAMAGE1",  makeIntValue(t.damage));

        ad->content.pushNew(new IntHash(hd));
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}

/******************************* BeamPacker ******************************/

/*
  Beam().COST.D                           beam[].COST->D
  Beam().COST.M                           beam[].COST->M
  Beam().COST.MC                          beam[].COST->MC
  Beam().COST.STR                         -
  Beam().COST.T                           beam[].COST->T
  Beam().DAMAGE                           beam[].DAMAGE
  Beam().ID                               -
  Beam().KILL                             beam[].KILL
  Beam().MASS                             beam[].MASS
  Beam().NAME                             beam[].NAME
  Beam().NAME.SHORT                       -
  Beam().TECH                             beam[].TECH
  Beam().TECH.BEAM                        -
*/

string_t
BeamPacker::pack(IntValue* data)
{
    string_t result;
    for (uint32_t i = 0, n = getArraySize(data); i < n; ++i) {
        IntValue* p = getArrayElem(data, i);

        TBeam b;
        packString(b.name, sizeof(b.name), toStringResult(getHashElem(p, "NAME")));
        packCost(b.cost, getHashElem(p, "COST"));
        b.mass   = toIntResult(getHashElem(p, "MASS"));
        b.tech   = toIntResult(getHashElem(p, "TECH"));
        b.kill   = toIntResult(getHashElem(p, "KILL"));
        b.damage = toIntResult(getHashElem(p, "DAMAGE"));

        storeStructure(result, b);
    }
    return result;
}

IntValue*
BeamPacker::unpack(const string_t& data)
{
    Ptr<IntArrayData> ad = new IntArrayData();
    TBeam b;
    string_t::size_type pos = 0;
    while (getStructure(data, pos, b)) {
        Ptr<IntHashData> hd = new IntHashData();
        hd->setNew("NAME",   makeStringValue(unpackString(b.name, sizeof(b.name))));
        hd->setNew("COST",   unpackCost(b.cost));
        hd->setNew("MASS",   makeIntValue(b.mass));
        hd->setNew("TECH",   makeIntValue(b.tech));
        hd->setNew("KILL",   makeIntValue(b.kill));
        hd->setNew("DAMAGE", makeIntValue(b.damage));

        ad->content.pushNew(new IntHash(hd));
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}

/******************************* HullPacker ******************************/

/*
  Hull().BEAM.MAX                         hullX.BEAM.MAX
  Hull().CARGO.MAX                        hullX.CARGO.MAX
  Hull().CARGO.MAXFUEL                    hullX.CARGO.MAXFUEL
  Hull().COST.D                           hullX.COST->D
  Hull().COST.M                           hullX.COST->M
  Hull().COST.MC                          hullX.COST->MC
  Hull().COST.STR                         -
  Hull().COST.T                           hullX.COST->T
  Hull().CREW.NORMAL                      hullX.CREW.NORMAL
  Hull().ENGINE.COUNT                     hullX.ENGINE.COUNT
  Hull().FIGHTER.BAYS                     hullX.FIGHTER.BAYS
  Hull().ID                               -
  Hull().IMAGE                            hullX.IMAGE
  Hull().IMAGE$                           -
  Hull().MASS                             hullX.MASS
  Hull().NAME                             hullX.NAME
  Hull().NAME.SHORT                       -
  Hull().SPECIAL                          -
  Hull().TECH                             hullX.TECH
  Hull().TECH.HULL                        -
  Hull().TORP.LMAX                        hullX.TORP.LMAX
*/

string_t
HullPacker::pack(IntValue* data)
{
    string_t result;
    for (uint32_t i = 0, n = getArraySize(data); i < n; ++i) {
        IntValue* p = getArrayElem(data, i);
        IntValue* pc = getHashElem(p, "COST");

        THull h;
        packString(h.name, sizeof(h.name), toStringResult(getHashElem(p, "NAME")));
        h.picture     = toIntResult(getHashElem(p, "IMAGE"));
        h.pcc_picture = 1;
        h.t           = toIntResult(getHashElem(pc, "T"));
        h.d           = toIntResult(getHashElem(pc, "D"));
        h.m           = toIntResult(getHashElem(pc, "M"));
        h.fuel        = toIntResult(getHashElem(p, "CARGO.MAXFUEL"));
        h.crew        = toIntResult(getHashElem(p, "CREW.NORMAL"));
        h.engines     = toIntResult(getHashElem(p, "ENGINE.COUNT"));
        h.mass        = toIntResult(getHashElem(p, "MASS"));
        h.tech        = toIntResult(getHashElem(p, "TECH"));
        h.cargo       = toIntResult(getHashElem(p, "CARGO.MAX"));
        h.bays        = toIntResult(getHashElem(p, "FIGHTER.BAYS"));
        h.launchers   = toIntResult(getHashElem(p, "TORP.LMAX"));
        h.beams       = toIntResult(getHashElem(p, "BEAM.MAX"));
        h.mc          = toIntResult(getHashElem(pc, "MC"));
        storeStructure(result, h);
    }
    return result;
}

IntValue*
HullPacker::unpack(const string_t& data)
{
    Ptr<IntArrayData> ad = new IntArrayData();
    THull h;
    string_t::size_type pos = 0;
    int hull_id = 0;
    while (getStructure(data, pos, h)) {
        /* Remap picture numbers. It's ugly to do this here, but this makes it
           somehow consistent with c2server. */
        ++hull_id;
        int pic_id = hull_id == 104 ? 152 : hull_id == 105 ? 153 : h.picture;

        TCost c;
        c.mc = h.mc;
        c.t  = h.t;
        c.d  = h.d;
        c.m  = h.m;

        Ptr<IntHashData> hd = new IntHashData();
        hd->setNew("NAME",          makeStringValue(unpackString(h.name, sizeof(h.name))));
        hd->setNew("IMAGE",         makeIntValue(pic_id));
        hd->setNew("COST",          unpackCost(c));
        hd->setNew("CARGO.MAXFUEL", makeIntValue(h.fuel));
        hd->setNew("CREW.NORMAL",   makeIntValue(h.crew));
        hd->setNew("ENGINE.COUNT",  makeIntValue(h.engines));
        hd->setNew("MASS",          makeIntValue(h.mass));
        hd->setNew("TECH",          makeIntValue(h.tech));
        hd->setNew("CARGO.MAX",     makeIntValue(h.cargo));
        hd->setNew("FIGHTER.BAYS",  makeIntValue(h.bays));
        hd->setNew("TORP.LMAX",     makeIntValue(h.launchers));
        hd->setNew("BEAM.MAX",      makeIntValue(h.beams));

        ad->content.pushNew(new IntHash(hd));
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}

/***************************** TruehullPacker ****************************/

string_t
TruehullPacker::pack(IntValue* data)
{
    TTruehull th;
    for (int pl = 0; pl < NUM_PLAYERS; ++pl) {
        for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i) {
            th.hulls[pl][i] = toIntResult(getArrayElem(getArrayElem(data, pl), i));
        }
    }
    string_t result;
    storeStructure(result, th);
    return result;
}

IntValue*
TruehullPacker::unpack(const string_t& data)
{
    Ptr<IntArrayData> ad = new IntArrayData();
    TTruehull th;
    string_t::size_type pos = 0;
    if (getStructure(data, pos, th)) {
        for (int pl = 0; pl < NUM_PLAYERS; ++pl) {
            Ptr<IntArrayData> ad2 = new IntArrayData();
            for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i) {
                ad2->content.pushNew(makeIntValue(th.hulls[pl][i]));
            }
            ad2->addDimension(ad2->content.size());
            ad->content.pushNew(new IntArray(ad2));
        }
    }
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}
