/**
  *  \file planetscentral/host/rank.cc
  *  \brief Host: Ranks
  *
  *  This is separate from victory.cc for testability.
  */

#include "planetscentral/host/rank.h"
#include "util/misc.h"


RefereeFileReader::RefereeFileReader()
    : end(false)
{
    initRanks(ranks);
}

void
RefereeFileReader::processLine(string_t line)
{
    /* Comment or blank? */
    string_t::size_type n = line.find_first_not_of(" \t");
    if (n >= line.size() || line[n] == '#' || line[n] == ';') {
        return;
    }

    /* Must be an assignment */
    string_t::size_type p = line.find('=');
    if (p == line.npos) {
        // FIXME: log it?
        return;
    }

    string_t key(strUCase(strTrim(line.substr(n, p-n))));
    string_t value(strTrim(line.substr(p+1)));
    int32_t player, val;
    if (key == "END" && stringValue(value, val)) {
        end = (val != 0);
    } else if (key.size() > 4
               && key.compare(0, 4, "RANK", 4) == 0
               && stringValue(key.substr(4), player)
               && player > 0
               && player <= NUM_PLAYERS)
    {
        if (stringValue(value, val)) {
            ranks[player-1] = val;
        }
    } else {
        // FIXME: "turns"
        // FIXME: warning?
    }
}



/** Initialize ranks.
    \param ranks [out] Output
    \param value [in] Value to initialize with */
void
initRanks(Rank_t& ranks, const int32_t value)
{
    for (int i = 0; i < NUM_PLAYERS; ++i) {
        ranks[i] = value;
    }
}

/** Compact ranks.
    \param dest    [out] Output ranks
    \param rank    [in]  Input ranks, smaller is better
    \param score   [in]  Input scores, used as tie-breaker for equal input ranks, bigger is better
    \param players [in]  Mask for slots used in either Rank_t */
void
compactRanks(Rank_t& dest, const Rank_t& rank, const Rank_t& score, GPlayerSet players)
{
    int place = 1;
    while (1) {
        /* Locate best-possible entry */
        GPlayerSet thisSlots;          // players on this rank
        int32_t    thisRank = 0;       // current value from rank
        int32_t    thisScore = 0;      // current value from score
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            if (players.contains(i)) {
                if (thisSlots.empty()
                    || rank[i-1] < thisRank
                    || (rank[i-1] == thisRank && score[i-1] > thisScore))
                {
                    thisSlots = GPlayerSet(i);
                    thisRank  = rank[i-1];
                    thisScore = score[i-1];
                } else if (rank[i-1] == thisRank && score[i-1] == thisScore) {
                    thisSlots += i;
                }
            }
        }

        /* Termination condition */
        if (thisSlots.empty()) {
            break;
        }

        /* Mark it */
        players -= thisSlots;
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            if (thisSlots.contains(i)) {
                dest[i-1] = place;
            }
        }
        ++place;
    }
}

#ifdef TEST
/*
  g++-3.3 -DTEST -Icpluslib -I. planetscentral/host/rank.cc
 */
#include <cassert>
#include <cstdio>
int main()
{
    {
        // Plain score ranking
        // Scores          #2    #4    #3    #5    #4    #1    #7    #6    #8    #7    #9
        Rank_t scores = { -500, -400, -450, -300, -400, -600, -200, -250, -100, -200, -50 };
        Rank_t null; initRanks(null);
        Rank_t result;
        compactRanks(result, scores, null, GPlayerSet::allUpTo(NUM_PLAYERS));
        assert(result[0]  == 2);
        assert(result[1]  == 4);
        assert(result[2]  == 3);
        assert(result[3]  == 5);
        assert(result[4]  == 4);
        assert(result[5]  == 1);
        assert(result[6]  == 7);
        assert(result[7]  == 6);
        assert(result[8]  == 8);
        assert(result[9]  == 7);
        assert(result[10] == 9);
    }

    {
        // Plain score ranking, with highest scores not playong
        // Scores            -   #2    #1    #3    #2     -    #5    #4    #6    #5    #7
        Rank_t scores = { -500, -400, -450, -300, -400, -600, -200, -250, -100, -200, -50 };
        Rank_t null; initRanks(null);
        Rank_t result;
        compactRanks(result, scores, null, GPlayerSet::allUpTo(NUM_PLAYERS) - 6 - 1);
        assert(result[1]  == 2);
        assert(result[2]  == 1);
        assert(result[3]  == 3);
        assert(result[4]  == 2);
        assert(result[6]  == 5);
        assert(result[7]  == 4);
        assert(result[8]  == 6);
        assert(result[9]  == 5);
        assert(result[10] == 7);
    }

    {
        // Turns-over-limit scoring
        // Scores         #4   #5    #3   #2  #5   #6   #7   #5    #1   #8   #1
        Rank_t turns  = { -1,  0,    -2,  -3,   0,   0,   0,   0,  -5,   0,  -5 };
        Rank_t scores = { 500, 400, 500, 500, 400, 300, 200, 400, 500, 100, 500 };
        Rank_t result;
        compactRanks(result, turns, scores, GPlayerSet::allUpTo(NUM_PLAYERS));
        assert(result[0]  == 4);
        assert(result[1]  == 5);
        assert(result[2]  == 3);
        assert(result[3]  == 2);
        assert(result[4]  == 5);
        assert(result[5]  == 6);
        assert(result[6]  == 7);
        assert(result[7]  == 5);
        assert(result[8]  == 1);
        assert(result[9]  == 8);
        assert(result[10] == 1);
    }

    {
        // Turns-over-limit scoring
        // Scores         #5   #6    #4   #3  #6   #7   #8   #6    #1   #9   #2
        Rank_t turns  = { -1,  0,    -2,  -3,   0,   0,   0,   0,  -5,   0,  -5 };
        Rank_t scores = { 500, 400, 500, 500, 400, 300, 200, 400, 501, 100, 500 };
        Rank_t result;
        compactRanks(result, turns, scores, GPlayerSet::allUpTo(NUM_PLAYERS));
        assert(result[0]  == 5);
        assert(result[1]  == 6);
        assert(result[2]  == 4);
        assert(result[3]  == 3);
        assert(result[4]  == 6);
        assert(result[5]  == 7);
        assert(result[6]  == 8);
        assert(result[7]  == 6);
        assert(result[8]  == 1);
        assert(result[9]  == 9);
        assert(result[10] == 2);
    }

    {
        // Ranking with a partial c2ref.txt file
        Rank_t ranks; initRanks(ranks);
        ranks[5] = 1;
        ranks[3] = 5;
        ranks[9] = 10;
        Rank_t scores; initRanks(scores);
        Rank_t result;
        compactRanks(result, ranks, scores, GPlayerSet::allUpTo(NUM_PLAYERS));
        assert(result[0]  == 4);
        assert(result[1]  == 4);
        assert(result[2]  == 4);
        assert(result[3]  == 2);
        assert(result[4]  == 4);
        assert(result[5]  == 1);
        assert(result[6]  == 4);
        assert(result[7]  == 4);
        assert(result[8]  == 4);
        assert(result[9]  == 3);
        assert(result[10] == 4);
    }

    {
        // Ranking with a partial c2ref.txt file and not all players participating
        // This is missing player 4 (ranks[3]).
        Rank_t ranks; initRanks(ranks);
        ranks[5] = 1;
        ranks[3] = 5;
        ranks[9] = 10;
        Rank_t scores; initRanks(scores);
        Rank_t result;
        compactRanks(result, ranks, scores, GPlayerSet::allUpTo(NUM_PLAYERS) - 4);
        assert(result[0]  == 3);
        assert(result[1]  == 3);
        assert(result[2]  == 3);
        assert(result[4]  == 3);
        assert(result[5]  == 1);
        assert(result[6]  == 3);
        assert(result[7]  == 3);
        assert(result[8]  == 3);
        assert(result[9]  == 2);
        assert(result[10] == 3);
    }

    {
        // Referee file
        RefereeFileReader rdr;
        assert(!rdr.end);

        rdr.processLine("rank1=99");
        rdr.processLine("rank2=98");
        rdr.processLine("rank3=97");
        rdr.processLine("rank4=96");
        rdr.processLine("RANK5=95");
        rdr.processLine("Rank6=94");
        rdr.processLine("rank7=93");
        rdr.processLine("RANK8=92");
        rdr.processLine("rank9=91");
        rdr.processLine("rank10=90");
        rdr.processLine("rank11=89");

        rdr.processLine("rank0=42");
        assert(rdr.ranks[0] == 99);
        assert(rdr.ranks[1] == 98);
        assert(rdr.ranks[2] == 97);
        assert(rdr.ranks[3] == 96);
        assert(rdr.ranks[4] == 95);
        assert(rdr.ranks[5] == 94);
        assert(rdr.ranks[6] == 93);
        assert(rdr.ranks[7] == 92);
        assert(rdr.ranks[8] == 91);
        assert(rdr.ranks[9] == 90);
        assert(rdr.ranks[10] == 89);
        assert(!rdr.end);

        rdr.processLine("end=1");
        assert(rdr.end);
    }

    std::puts("ok");
}
#endif
