/**
  *  \file planetscentral/nntp/nntp.cc
  *  \brief NNTP Server Main Program
  */

#include <sys/socket.h>
#include <netdb.h>
#include "planetscentral/nntp/nntp.h"
#include <stdio.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "cpluslib/string.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/osutil.h"
#include "server/config.h"
#include "server/net.h"
#include "planetscentral/nntp/iothread.h"
#include "planetscentral/nntp/nntpworker.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "1119";

Mutex output_mutex;
DbClient talk_connection;
ReconnectingConnection talk_rc(talk_connection, output_mutex, "127.0.0.1", "5555");
string_t www_prefix;

static void
processConfig(string_t key, string_t value)
{
    if (key == "NNTP.HOST") {
        arg_listen_host = value;
    } else if (key == "NNTP.PORT") {
        arg_listen_port = value;
    } else if (key == "TALK.HOST") {
        talk_rc.setHost(value);
    } else if (key == "TALK.PORT") {
        talk_rc.setPort(value);
    } else if (key == "TALK.WWWROOT") {
        www_prefix = value;
    } else {
        // ignore
    }
}


int main(int /*argc*/, char** /*argv*/)
{
    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Configuration */
    loadConfiguration(processConfig);

    /* Connect to server */
    try {
        talk_rc.connectRetry();
    }
    catch (std::exception& e) {
        fprintf(stderr, "unable to connect to servers: %s\n", e.what());
        return 1;
    }

    /* Listen */
    int server_socket;
    if (char* s = getenv("C2SOCKET")) {
        char* result;
        server_socket = strtol(s, &result, 0);
        if (result == 0 || *result != '\0' || server_socket < 0) {
            fprintf(stderr, "[main] unable to parse provided socket '%s'\n", s);
            return 1;
        }
    } else {
        server_socket = Bind().withAddress(arg_listen_host, arg_listen_port);
        if (server_socket < 0) {
            fprintf(stderr, "[main] unable to bind socket\n");
            return 1;
        }
    }

    /* OK, main loop */
    printf("[main] listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(server_socket, F_SETFD, FD_CLOEXEC);
    listen(server_socket, 10);

    /* Start an IO thread */
    IOThread io;
    io.start();

    /* Accept connections */
    int errorCounter = 0;
    while (1) {
        sockaddr_storage addr;
        socklen_t len = sizeof(addr);
        int fd = accept(server_socket, (sockaddr*) &addr, &len);
        if (fd < 0) {
            MutexGuard g(output_mutex);
            printf("[main] accept error %d (%s)\n", errno, strerror(errno));
            ++errorCounter;
            if (errorCounter > 100) {
                break;
            }
            if (errorCounter > 10) {
                sleep(5);
            }
        } else {
            std::auto_ptr<NntpWorker> worker(new NntpWorker());
            char name[1024], serv[32];
            if (getnameinfo((sockaddr*) &addr, len, name, sizeof(name), serv, sizeof(serv), NI_NUMERICHOST | NI_NUMERICSERV) == 0) {
                MutexGuard g(output_mutex);
                printf("[main] [fd:%d] [id:%d] connection accepted from '%s:%s'\n", fd, worker->getId(), name, serv);
            } else {
                MutexGuard g(output_mutex);
                printf("[main] [fd:%d] [id:%d] connection accepted (name resolution failed)\n", fd, worker->getId());
            }
            io.addWorker(fd, worker.release());
        }
    }
    close(server_socket);
}
